#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ScalerBase.

Author: Andrew Lee
"""
import pytest
import re

from pyomo.environ import ConcreteModel, Constraint, Set, Suffix, Var
from pyomo.common.config import ConfigDict

from idaes.core.scaling.scaling_base import ScalerBase
import idaes.logger as idaeslog


@pytest.fixture
def model():
    m = ConcreteModel()

    m.s = Set(initialize=[1, 2, 3, 4])

    m.v = Var(m.s)

    def c_rule(b, i):
        return b.v[i] == i

    m.c = Constraint(m.s, rule=c_rule)

    m.scaling_factor = Suffix(direction=Suffix.EXPORT)

    m.scaling_factor[m.v[1]] = 1
    m.scaling_factor[m.v[2]] = 2
    m.scaling_factor[m.v[3]] = 3
    m.scaling_factor[m.v[4]] = 4
    m.scaling_factor[m.c[1]] = 11
    m.scaling_factor[m.c[2]] = 21
    m.scaling_factor[m.c[3]] = 31
    m.scaling_factor[m.c[4]] = 41

    return m


class TestScalerBase:
    @pytest.mark.unit
    def test_init(self):
        sb = ScalerBase()

        assert isinstance(sb.config, ConfigDict)
        assert sb.config.zero_tolerance == 1e-12
        assert sb.config.max_variable_scaling_factor == 1e10
        assert sb.config.min_variable_scaling_factor == 1e-10
        assert sb.config.max_constraint_scaling_factor == 1e10
        assert sb.config.min_constraint_scaling_factor == 1e-10
        assert not sb.config.overwrite

    @pytest.mark.unit
    def test_get_scaling_factor(self, model):
        sb = ScalerBase()

        assert sb.get_scaling_factor(model.v[1]) == 1

    @pytest.mark.unit
    def test_set_scaling_factor(self, model):
        sb = ScalerBase()

        sb._set_scaling_factor(
            component=model.v[1],
            component_type="variable",
            scaling_factor=42,
            overwrite=False,
        )
        sb._set_scaling_factor(
            component=model.c[1],
            component_type="constraint",
            scaling_factor=42,
            overwrite=False,
        )
        assert model.scaling_factor[model.v[1]] == 1  # Overwrite = False, no change
        assert model.scaling_factor[model.c[1]] == 11  # Overwrite = False, no change

        sb._set_scaling_factor(
            component=model.v[1],
            component_type="variable",
            scaling_factor=42,
            overwrite=True,
        )
        sb._set_scaling_factor(
            component=model.c[1],
            component_type="constraint",
            scaling_factor=42,
            overwrite=True,
        )
        assert model.scaling_factor[model.v[1]] == 42
        assert model.scaling_factor[model.c[1]] == 42

    @pytest.mark.unit
    def test_set_scaling_factor_overwrite_default(self, model):
        sb = ScalerBase(overwrite=False)

        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=42
        )
        sb._set_scaling_factor(
            component=model.c[1], component_type="constraint", scaling_factor=42
        )
        assert model.scaling_factor[model.v[1]] == 1  # Overwrite = False, no change
        assert model.scaling_factor[model.c[1]] == 11  # Overwrite = False, no change

        # Change default overwrite setting
        sb.config.overwrite = True
        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=42
        )
        sb._set_scaling_factor(
            component=model.c[1], component_type="constraint", scaling_factor=42
        )
        assert model.scaling_factor[model.v[1]] == 42
        assert model.scaling_factor[model.c[1]] == 42

    @pytest.mark.unit
    def test_set_scaling_factor_invalid_component_type(self, model):
        sb = ScalerBase()

        with pytest.raises(
            ValueError,
            match="Invalid value for component_type.",
        ):
            sb._set_scaling_factor(
                component=model.v[1], component_type="foo", scaling_factor="bar"
            )

    @pytest.mark.unit
    def test_set_scaling_factor_max_limit(self, model, caplog):
        sb = ScalerBase(
            overwrite=True,
            max_variable_scaling_factor=100,
            max_constraint_scaling_factor=200,
        )

        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )
        # Below max var limit - value set as given
        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=42
        )
        assert model.scaling_factor[model.v[1]] == 42

        # Above max var limit - value limited to max
        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=150
        )
        assert model.scaling_factor[model.v[1]] == 100
        assert (
            "Scaling factor for v[1] limited by maximum value (max_sf: 100.0 < sf: 150)"
            in caplog.text
        )

        # Above max var limit but below max con limit - value set as given
        sb._set_scaling_factor(
            component=model.v[1], component_type="constraint", scaling_factor=150
        )
        assert model.scaling_factor[model.v[1]] == 150

        # Above max con limit - value limited to max
        sb._set_scaling_factor(
            component=model.v[1], component_type="constraint", scaling_factor=250
        )
        assert model.scaling_factor[model.v[1]] == 200
        assert (
            "Scaling factor for v[1] limited by maximum value (max_sf: 200.0 < sf: 250)"
            in caplog.text
        )

    @pytest.mark.unit
    def test_set_scaling_factor_min_limit(self, model, caplog):
        sb = ScalerBase(
            overwrite=True,
            min_variable_scaling_factor=100,
            min_constraint_scaling_factor=200,
        )

        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        # Above both min limits - value set as given
        sb._set_scaling_factor(
            component=model.v[1], component_type="constraint", scaling_factor=400
        )
        assert model.scaling_factor[model.v[1]] == 400

        # Below con min limit - value limited
        sb._set_scaling_factor(
            component=model.v[1], component_type="constraint", scaling_factor=150
        )
        assert model.scaling_factor[model.v[1]] == 200
        assert (
            "Scaling factor for v[1] limited by minimum value (min_sf: 200.0 > sf: 150)"
            in caplog.text
        )

        # Below con min limit but above min var limit - value set as given
        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=150
        )
        assert model.scaling_factor[model.v[1]] == 150

        # Below both limited
        sb._set_scaling_factor(
            component=model.v[1], component_type="variable", scaling_factor=50
        )
        assert model.scaling_factor[model.v[1]] == 100
        assert (
            "Scaling factor for v[1] limited by minimum value (min_sf: 100.0 > sf: 50)"
            in caplog.text
        )

    @pytest.mark.unit
    def test_set_variable_scaling_factor(self, model, caplog):
        sb = ScalerBase(
            max_variable_scaling_factor=1e3,
            min_variable_scaling_factor=100,
            max_constraint_scaling_factor=2e3,
            min_constraint_scaling_factor=200,
            overwrite=True,
        )

        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Scaling factor within limits
        sb.set_variable_scaling_factor(model.v[1], 200)
        assert model.scaling_factor[model.v[1]] == 200

        # Too large
        sb.set_variable_scaling_factor(model.v[1], 2e3)
        assert model.scaling_factor[model.v[1]] == 1e3
        assert (
            "Scaling factor for v[1] limited by maximum value (max_sf: 1000.0 < sf: 2000.0)"
            in caplog.text
        )

        # Too small
        sb.set_variable_scaling_factor(model.v[1], 1)
        assert model.scaling_factor[model.v[1]] == 100
        assert (
            "Scaling factor for v[1] limited by minimum value (min_sf: 100.0 > sf: 1)"
            in caplog.text
        )

    @pytest.mark.unit
    def test_set_variable_scaling_factor_invalid_type(self, model):
        sb = ScalerBase()

        with pytest.raises(
            TypeError, match=re.escape("c[1] is not a variable (or is indexed).")
        ):
            sb.set_variable_scaling_factor(model.c[1], 200)

        with pytest.raises(
            TypeError, match=re.escape("v is not a variable (or is indexed).")
        ):
            sb.set_variable_scaling_factor(model.v, 200)

    @pytest.mark.unit
    def test_set_constraint_scaling_factor(self, model, caplog):
        sb = ScalerBase(
            max_variable_scaling_factor=1e3,
            min_variable_scaling_factor=100,
            max_constraint_scaling_factor=2e3,
            min_constraint_scaling_factor=200,
            overwrite=True,
        )

        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Scaling factor within limits
        sb.set_constraint_scaling_factor(model.c[1], 200)
        assert model.scaling_factor[model.c[1]] == 200

        # Too large
        sb.set_constraint_scaling_factor(model.c[1], 3e3)
        assert model.scaling_factor[model.c[1]] == 2e3
        assert (
            "Scaling factor for c[1] limited by maximum value (max_sf: 2000.0 < sf: 3000.0)"
            in caplog.text
        )

        # Too small
        sb.set_constraint_scaling_factor(model.c[1], 1)
        assert model.scaling_factor[model.c[1]] == 200
        assert (
            "Scaling factor for c[1] limited by minimum value (min_sf: 200.0 > sf: 1)"
            in caplog.text
        )

    @pytest.mark.unit
    def test_set_constraint_scaling_factor_invalid_type(self, model):
        sb = ScalerBase()

        with pytest.raises(
            TypeError, match=re.escape("v[1] is not a constraint (or is indexed).")
        ):
            sb.set_constraint_scaling_factor(model.v[1], 200)

        with pytest.raises(
            TypeError, match=re.escape("c is not a constraint (or is indexed).")
        ):
            sb.set_constraint_scaling_factor(model.c, 200)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Condenser model for distillation.

While the condenser model (both total and partial), is fairly simple, a major
portion of this code has gone into making this generic and be able to handle
different state variables and the associated splits.
"""

__author__ = "Jaffer Ghouse"

from enum import Enum
from pandas import DataFrame

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Port
from pyomo.environ import (
    Reference,
    Var,
    Constraint,
    value,
    Set,
    check_optimal_termination,
)

# Import IDAES cores
import idaes.logger as idaeslog
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MaterialBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import (
    ConfigurationError,
    InitializationError,
)
from idaes.core.solvers import get_solver
from idaes.models_extra.column_models.util import make_phase_split

_log = idaeslog.getLogger(__name__)


class CondenserType(Enum):
    """
    Enum for supported condenser types.
    """

    totalCondenser = 0
    partialCondenser = 1


class TemperatureSpec(Enum):
    """
    Enum for temperature specifications.
    """

    atBubblePoint = 0
    customTemperature = 1


@declare_process_block_class("Condenser")
class CondenserData(UnitModelBlockData):
    """
    Condenser unit for distillation model.
    Unit model to condense (total/partial) the vapor from the top tray of
    the distillation column.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "condenser_type",
        ConfigValue(
            default=CondenserType.totalCondenser,
            domain=In(CondenserType),
            description="Type of condenser flag",
            doc="""Indicates what type of condenser should be constructed,
**default** - CondenserType.totalCondenser.
**Valid values:** {
**CondenserType.totalCondenser** - Incoming vapor from top tray is condensed
to all liquid,
**CondenserType.partialCondenser** - Incoming vapor from top tray is
partially condensed to a vapor and liquid stream.}""",
        ),
    )
    CONFIG.declare(
        "temperature_spec",
        ConfigValue(
            default=None,
            domain=In(TemperatureSpec),
            description="Temperature spec for the condenser",
            doc="""Temperature specification for the condenser,
**default** - TemperatureSpec.none
**Valid values:** {
**TemperatureSpec.none** - No spec is selected,
**TemperatureSpec.atBubblePoint** - Condenser temperature set at
bubble point i.e. total condenser,
**TemperatureSpec.customTemperature** - Condenser temperature at
user specified temperature.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """Build the model.
        Args:
            None
        Returns:
            None
        """
        # Setup model build logger
        model_log = idaeslog.getModelLogger(self.name, tag="unit")

        # Call UnitModel.build to setup dynamics
        super(CondenserData, self).build()

        # Check config arguments
        if self.config.temperature_spec is None:
            raise ConfigurationError(
                "temperature_spec config argument "
                "has not been specified. Please select "
                "a valid option."
            )
        if (self.config.condenser_type == CondenserType.partialCondenser) and (
            self.config.temperature_spec == TemperatureSpec.atBubblePoint
        ):
            raise ConfigurationError(
                "condenser_type set to partial but "
                "temperature_spec set to atBubblePoint. "
                "Select customTemperature and specify "
                "outlet temperature."
            )

        # Add Control Volume for the condenser
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_phase_equilibrium=True
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_heat_transfer=True
        )

        # Note: No momentum balance added for the condenser as the condenser
        # outlet pressure is a spec set by the user.

        # Get liquid and vapor phase objects from the property package
        # to be used below. Avoids repetition.
        _liquid_list = []
        _vapor_list = []
        for p in self.control_volume.config.property_package.phase_list:
            pobj = self.control_volume.config.property_package.get_phase(p)
            if pobj.is_vapor_phase():
                _vapor_list.append(p)
            elif pobj.is_liquid_phase():
                _liquid_list.append(p)
            else:
                _liquid_list.append(p)
                model_log.warning(
                    "A non-liquid/non-vapor phase was detected but will "
                    "be treated as a liquid."
                )

        # Create a pyomo set for indexing purposes. This set is appended to
        # model otherwise results in an abstract set.
        self._liquid_set = Set(initialize=_liquid_list)
        self._vapor_set = Set(initialize=_vapor_list)

        self._make_ports()

        @self.Expression(doc="Split fraction calculation calculations")
        def reflux_split_fraction(self):
            return self.reflux_ratio * (1 + self.reflux_ratio) ** -1

        if self.config.condenser_type == CondenserType.totalCondenser:

            make_phase_split(
                self.control_volume,
                port=self.reflux,
                phase=self._liquid_set,
                side_sf=self.reflux_split_fraction,
                equipmentType=CondenserType.totalCondenser,
            )

            make_phase_split(
                self.control_volume,
                port=self.distillate,
                phase=self._liquid_set,
                side_sf=1 - self.reflux_split_fraction,
                equipmentType=CondenserType.totalCondenser,
            )

            if self.config.temperature_spec == TemperatureSpec.atBubblePoint:
                # Option 1: if true, condition for total condenser
                # (T_cond = T_bubble)
                # Option 2: if this is false, then user has selected
                # custom temperature spec and needs to fix an outlet
                # temperature.

                # Get index for bubble point temperature and and assume it
                # will have only a single phase equilibrium pair. This is to
                # support the generic property framework where the T_bubble
                # is indexed by the phases_in_equilibrium. In distillation,
                # the assumption is that there will only be a single pair
                # i.e. vap-liq.
                idx = next(
                    iter(
                        self.control_volume.properties_out[
                            self.flowsheet().time.first()
                        ].temperature_bubble
                    )
                )

                def rule_total_cond(self, t):
                    return (
                        self.control_volume.properties_out[t].temperature
                        == self.control_volume.properties_out[t].temperature_bubble[idx]
                    )

                self.eq_total_cond_spec = Constraint(
                    self.flowsheet().time, rule=rule_total_cond
                )

        else:

            make_phase_split(
                self.control_volume,
                port=self.reflux,
                phase=self._liquid_set,
                side_sf=self.reflux_split_fraction,
                equipmentType=CondenserType.partialCondenser,
            )

            make_phase_split(
                self.control_volume,
                port=self.distillate,
                phase=self._liquid_set,
                side_sf=1 - self.reflux_split_fraction,
                equipmentType=CondenserType.partialCondenser,
            )
            make_phase_split(
                self.control_volume,
                port=self.vapor_outlet,
                phase=self._vapor_set,
                # Split fraction set to 1 as all vapor from condenser
                # returns to column
                side_sf=1,
                equipmentType=CondenserType.partialCondenser,
            )

        # Add object reference to variables of the control volume
        # Reference to the heat duty
        self.heat_duty = Reference(self.control_volume.heat[:])

        self.condenser_pressure = Reference(
            self.control_volume.properties_out[:].pressure
        )

    def _make_ports(self):

        # Add Ports for the condenser
        # Inlet port (the vapor from the top tray)
        self.add_inlet_port()

        # Outlet ports that always exist irrespective of condenser type
        self.reflux = Port(
            noruleinit=True, doc="Reflux stream that is" " returned to the top tray."
        )
        self.distillate = Port(
            noruleinit=True, doc="Distillate stream that is" " the top product."
        )

        if self.config.condenser_type == CondenserType.partialCondenser:
            self.vapor_outlet = Port(
                noruleinit=True, doc="Vapor outlet port from a " "partial condenser"
            )
        # Add codnenser specific variables
        self.reflux_ratio = Var(initialize=0.5, doc="Reflux ratio for the condenser")

    def initialize(
        self, state_args=None, solver=None, optarg=None, outlvl=idaeslog.NOTSET
    ):

        # TODO: Fix the inlets to the condenser to the vapor flow from
        # the top tray or take it as an argument to this method.

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        if self.config.temperature_spec == TemperatureSpec.customTemperature:
            if degrees_of_freedom(self) != 0:
                raise ConfigurationError(
                    "Degrees of freedom is not 0 during initialization. "
                    "Check if outlet temperature has been fixed in addition "
                    "to the other inputs required as customTemperature was "
                    "selected for temperature_spec config argument."
                )

        solverobj = get_solver(solver, optarg)

        if state_args is None:
            state_args = {}
            state_dict = self.control_volume.properties_in[
                self.flowsheet().time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = value(state_dict[k][m])
                else:
                    state_args[k] = value(state_dict[k])

        if self.config.condenser_type == CondenserType.totalCondenser:
            self.eq_total_cond_spec.deactivate()

        # Initialize the inlet and outlet state blocks
        flags = self.control_volume.initialize(
            state_args=state_args,
            solver=solver,
            optarg=optarg,
            outlvl=outlvl,
            hold_state=True,
        )

        # Activate the total condenser spec
        if self.config.condenser_type == CondenserType.totalCondenser:
            self.eq_total_cond_spec.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self, tee=slc.tee)
        init_log.info("Initialization Complete, {}.".format(idaeslog.condition(res)))
        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        self.control_volume.release_state(flags=flags)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}

        if self.config.condenser_type == CondenserType.totalCondenser:
            stream_dict = {
                "Inlet": "inlet",
                "Reflux": "reflux",
                "Distillate": "distillate",
            }
        else:
            stream_dict = {
                "Inlet": "inlet",
                "Vapor Outlet": "vapor_outlet",
                "Reflux": "reflux",
                "Distillate": "distillate",
            }

        for n, v in stream_dict.items():
            port_obj = getattr(self, v)

            stream_attributes[n] = {}

            for k in port_obj.vars:
                for i in port_obj.vars[k].keys():
                    if isinstance(i, float):
                        stream_attributes[n][k] = value(port_obj.vars[k][time_point])
                    else:
                        if len(i) == 2:
                            kname = str(i[1])
                        else:
                            kname = str(i[1:])
                        stream_attributes[n][k + " " + kname] = value(
                            port_obj.vars[k][time_point, i[1:]]
                        )

        return DataFrame.from_dict(stream_attributes, orient="columns")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
This module contains miscellaneous utility functions for use in IDAES models.
"""
from enum import Enum
import sys

import pyomo.environ as pyo
from pyomo.common.config import ConfigBlock
from pyomo.core.expr.visitor import _ToStringVisitor
from pyomo.core.base.expression import ExpressionData

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


# Author: Andrew Lee
def add_object_reference(self, local_name, remote_object):
    """
    Method to create a reference in the local model to a remote Pyomo object.
    This method should only be used where Pyomo Reference objects are not
    suitable (such as for referencing scalar Pyomo objects where the None
    index is undesirable).

    Args:
        local_name : name to use for local reference (str)
        remote_object : object to make a reference to

    Returns:
        None
    """
    try:
        object.__setattr__(self, local_name, remote_object)
    except AttributeError:
        raise AttributeError(
            "{} failed to construct reference to {} - remote "
            "object does not exist.".format(self.name, remote_object)
        )


# Author: Jaffer Ghouse
def extract_data(data_dict):
    """
    General method that returns a rule to extract data from a python
    dictionary. This method allows the param block to have a database for
    a parameter but extract a subset of this data to initialize a Pyomo
    param object.
    """

    def _rule_initialize(m, *args):
        if len(args) > 1:
            return data_dict[args]
        else:
            return data_dict[args[0]]

    return _rule_initialize


def set_param_from_config(b, param, config=None, index=None):
    """
    Utility method to set parameter value from a config block. This allows for
    converting units if required. This method directly sets the value of the
    parameter.

    This method supports three forms for defining the parameter value:

    1. a 2-tuple of the form (value, units) where units are the units that
    the value are defined in
    2. a float where the float is assumed to be the value of the parameter
    value in the base units of the property package
    3. a Python Class which has a get_parameter_value method which will
    return a 2-tuple of (value, units) based on a lookup of the parameter name

    Args:
        b - block on which parameter and config block are defined
        param - name of parameter as str. Used to find param and config arg
        units - units of param object (used if conversion required)
        config - (optional) config block to get parameter data from. If
                unset, assumes b.config.
        index - (optional) used for pure component properties where a single
                property may have multiple parameters associated with it.

    Returns:
        None
    """
    if config is None:
        try:
            config = b.config
        except AttributeError:
            raise AttributeError(
                "{} - set_param_from_config method was not provided with a "
                "config argument, but no default Config block exists. Please "
                "specify the Config block to use via the config argument.".format(
                    b.name
                )
            )

    # Check that config is an instance of a Config Block
    if not isinstance(config, ConfigBlock):
        raise TypeError(
            "{} - set_param_from_config - config argument provided is not an "
            "instance of a Config Block.".format(b.name)
        )

    if index is None:
        try:
            param_obj = getattr(b, param)
        except AttributeError:
            raise AttributeError(
                "{} - set_param_from_config method was provided with param "
                "argument {}, but no attribute of that name exists.".format(
                    b.name, param
                )
            )

        try:
            p_data = config.parameter_data[param]
        except (KeyError, AttributeError):
            raise KeyError(
                "{} - set_param_from_config method was provided with param "
                "argument {}, but the config block does not contain a "
                "value for this parameter.".format(b.name, param)
            )
    else:
        try:
            param_obj = getattr(b, param + "_" + index)
        except AttributeError:
            raise AttributeError(
                f"{b.name} - set_param_from_config method was provided with param and"
                f" index arguments {param} {index}, but no attribute with that "
                f"combination ({param}_{index}) exists."
            )

        try:
            p_data = config.parameter_data[param][index]
        except (KeyError, AttributeError):
            raise KeyError(
                "{} - set_param_from_config method was provided with param and"
                " index arguments {} {}, but the config block does not contain"
                " a value for this parameter and index.".format(b.name, param, index)
            )

    # Check to see if p_data is callable, and if so, try to call the
    # get_parameter_value method to get 2-tuple
    if hasattr(p_data, "get_parameter_value"):
        p_data = p_data.get_parameter_value(b.local_name, param)

    if isinstance(p_data, tuple):
        if p_data[1] is None:
            p_val = p_data[0] * pyo.units.dimensionless
        else:
            p_val = p_data[0] * p_data[1]
    else:
        _log.debug(
            "{} no units provided for parameter {} - assuming default "
            "units".format(b.name, param)
        )
        p_val = p_data

    param_obj.set_value(p_val)


class StrEnum(str, Enum):
    """
    Multiple inheritance string-Enum for representing Enums with string values
    """

    def __str__(self):
        return str(self.value)


class _ToExprStringVisitor(_ToStringVisitor):
    """
    Derived version of the Pyomo _ToStringVisitor class
    which checks for named Expressions in the expression tree
    and prints their name instead of expanding the expression tree.
    """

    def visiting_potential_leaf(self, node):
        # Check if node is a named Expression
        if isinstance(node, ExpressionData):
            # If it is, return the name of the Expression
            return True, node.name

        # Otherwise, continue descending as normal
        return super().visiting_potential_leaf(node)


def compact_expression_to_string(expr):
    """Return a compact string representation of an expression.

    Unlike the normal Pyomo string representations, this function will
    identify any Expressions that appear within the expression tree and
    print the name of these rather than expanding the expression tree.

    Args:
        expr: ExpressionBase. The root node of an expression tree.

    Returns:
        A string representation for the expression.
    """
    # Create and execute the visitor pattern
    #
    visitor = _ToExprStringVisitor(verbose=False, smap=None)
    return visitor.dfs_postorder_stack(expr)


def print_compact_form(expr, stream=None):
    """
    Writes a compact string representation of the given component or
    expression to the specified stream.

    Unlike the normal Pyomo string representations, this function will
    identify any Expressions that appear within the expression tree and
    print the name of these rather than expanding the expression tree.

    Args:
        expr: component or expression to print
        stream: StringIO object to write to. Default is stdout.

    Returns:
        None
    """
    if stream is None:
        stream = sys.stdout

    if hasattr(expr, "expr"):
        # We have a Constraint or Expression
        # We want to print the expression, not the object name
        expr = expr.expr

    stream.write(compact_expression_to_string(expr))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""This module provides functions to register and retrieve component information
"""

_components = {}


class _ComponentStruct(object):
    """Component registry entry structure"""

    def __init__(
        self,
        viscosity=False,
        thermal_conductivity=False,
        surface_tension=False,
        eos_ref=None,
        viscosity_ref=None,
        thermal_conductivity_ref=None,
        surface_tension_ref=None,
    ):
        """Create a component registry object.

        Args:
            viscosity (bool): True if viscosity model available
            thermal_conductivity (bool): True if thermal conductivity model available
            surface_tension (bool): True if surface tension model available
            eos_ref (str|list|None): Reference for equation of state model
            viscosity_ref (str|list|None): Reference for viscosity model
            thermal_conductivity_ref (str|list|None): Reference for thermal conductivity model
            surface_tension_ref (str|list|None): Reference for surface tension model

        Returns:
            _ComponentStruct"""
        # these are true if external functions to calculate are available
        self.viscosity = viscosity
        self.thermal_conductivity = thermal_conductivity
        self.surface_tension = surface_tension
        if isinstance(eos_ref, str) or eos_ref is None:
            self.eos_ref = eos_ref
        else:
            self.eos_ref = "\n".join(eos_ref)
        if isinstance(viscosity_ref, str) or viscosity_ref is None:
            self.viscosity_ref = viscosity_ref
        else:
            self.viscosity_ref = "\n".join(viscosity_ref)
        if (
            isinstance(thermal_conductivity_ref, str)
            or thermal_conductivity_ref is None
        ):
            self.thermal_conductivity_ref = thermal_conductivity_ref
        else:
            self.thermal_conductivity_ref = "\n".join(thermal_conductivity_ref)
        if isinstance(surface_tension_ref, str) or surface_tension_ref is None:
            self.surface_tension_ref = surface_tension_ref
        else:
            self.surface_tension_ref = "\n".join(surface_tension_ref)


def register_helmholtz_component(
    comp_str,
    viscosity=False,
    thermal_conductivity=False,
    surface_tension=False,
    eos_ref=None,
    viscosity_ref=None,
    thermal_conductivity_ref=None,
    surface_tension_ref=None,
):
    """Add a component to the registry

    Args:
        viscosity (bool): True if viscosity model available
        thermal_conductivity (bool): True if thermal conductivity model available
        surface_tension (bool): True if surface tension model available
        eos_ref (str|list|None): Reference for equation of state model
        viscosity_ref (str|list|None): Reference for viscosity model
        thermal_conductivity_ref (str|list|None): Reference for thermal conductivity model
        surface_tension_ref (str|list|None): Reference for surface tension model

    Returns:
        None
    """
    comp_str = comp_str.lower()
    _components[comp_str] = _ComponentStruct(
        viscosity=viscosity,
        thermal_conductivity=thermal_conductivity,
        surface_tension=surface_tension,
        eos_ref=eos_ref,
        viscosity_ref=viscosity_ref,
        thermal_conductivity_ref=thermal_conductivity_ref,
        surface_tension_ref=surface_tension_ref,
    )


def remove_component(comp_str):
    """Remove a component from the registry. If the component isn't
    registered, raise KeyError.

    Args:
        comp_str (str): Component to remove

    Returns:
        None

    Raises:
        KeyError: if the component isn't registered raise a KeyError
    """
    comp_str = comp_str.lower()
    del _components[comp_str]


def clear_component_registry():
    """Remove all components from registry"""
    _components.clear()


def registered_components():
    """Return a list of registered components"""
    return list(_components.keys())


def viscosity_available(comp_str):
    """Return whether a viscosity model is available for a component.

    Args:
        comp_str (str): component string

    Returns:
        bool: True if viscosity model is available, False if not
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].viscosity
    return False


def thermal_conductivity_available(comp_str):
    """Return whether a thermal conductivity model is available for a component.

    Args:
        comp_str (str): component string

    Returns:
        bool: True if thermal conductivity model is available, False if not
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].thermal_conductivity
    return False


def surface_tension_available(comp_str):
    """Return whether a surface tension model is available for a component.

    Args:
        comp_str (str): component string

    Returns:
        bool: True if surface tension model is available, False if not
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].surface_tension
    return False


def component_registered(comp_str):
    """Return whether a component is registered.

    Args:
        comp_str (str): component string

    Returns:
        bool: True if component is available, False if not
    """
    comp_str = comp_str.lower()
    return comp_str in _components


def eos_reference(comp_str):
    """Return the equation of state reference or None if
    not available

    Args:
        comp_str (str): component string

    Returns:
        (str|None): Equation of state reference
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].eos_ref
    return None


def viscosity_reference(comp_str):
    """Return the viscosity reference or None if
    not available

    Args:
        comp_str (str): component string

    Returns:
        (str|None): Viscosity reference
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].viscosity_ref
    return None


def thermal_conductivity_reference(comp_str):
    """Return the thermal conductivity reference or None if
    not available

    Args:
        comp_str (str): component string

    Returns:
        (str|None): Thermal conductivity reference
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].thermal_conductivity_ref
    return None


def surface_tension_reference(comp_str):
    """Return the surface tension reference or None if
    not available

    Args:
        comp_str (str): component string

    Returns:
        (str|None): Surface tension reference
    """
    comp_str = comp_str.lower()
    if component_registered(comp_str):
        return _components[comp_str].surface_tension_ref
    return None



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import sys
import os
import io
from unittest.mock import patch

sys.path.append(os.path.abspath(".."))  # current folder is ~/tests
from idaes.core.surrogate.pysmo.kriging import KrigingModel
import numpy as np
import pandas as pd
import pytest


@pytest.mark.usefixtures("run_class_in_tmp_path")
class TestKrigingModel:
    y = np.array(
        [
            [i, j, ((i + 1) ** 2) + ((j + 1) ** 2)]
            for i in np.linspace(0, 10, 21)
            for j in np.linspace(0, 10, 21)
        ]
    )
    full_data = {"x1": y[:, 0], "x2": y[:, 1], "y": y[:, 2]}
    training_data = [
        [i, j, ((i + 1) ** 2) + ((j + 1) ** 2)]
        for i in np.linspace(0, 10, 5)
        for j in np.linspace(0, 10, 5)
    ]
    test_data = [[i, (i + 1) ** 2] for i in range(10)]
    test_data_large = [[i, (i + 1) ** 2] for i in range(200)]
    test_data_1d = [[(i + 1) ** 2] for i in range(10)]
    test_data_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(10)]
    sample_points = [[i, (i + 1) ** 2] for i in range(8)]
    sample_points_large = [[i, (i + 1) ** 2] for i in range(100)]
    sample_points_1d = [[(i + 1) ** 2] for i in range(8)]
    sample_points_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(8)]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__01(self, array_type):
        input_array = array_type(self.test_data)
        KrigingClass = KrigingModel(input_array)
        assert KrigingClass.num_grads == True
        assert KrigingClass.regularization == True

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [list])
    def test__init__02(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(ValueError):
            KrigingClass = KrigingModel(input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__03(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(Exception):
            KrigingClass = KrigingModel(input_array, numerical_gradients=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__04(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(Exception):
            KrigingClass = KrigingModel(input_array, regularization=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__05(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(Exception):
            KrigingClass = KrigingModel(input_array, overwrite=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__06(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(Exception):
            KrigingClass = KrigingModel(input_array, fname="solution.pkl")

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__07(self, array_type):
        input_array = array_type(self.test_data)
        with pytest.raises(Exception):
            KrigingClass = KrigingModel(input_array, fname=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    @pytest.mark.filterwarnings(
        "ignore:invalid value encountered in log:RuntimeWarning"
    )
    def test__init__08(self, array_type):
        input_array = array_type(self.test_data)
        file_name = "test_filename.pickle"
        KrigingClass1 = KrigingModel(input_array, fname=file_name, overwrite=True)
        results = KrigingClass1.training()
        KrigingClass2 = KrigingModel(input_array, fname=file_name, overwrite=True)
        assert KrigingClass1.filename == KrigingClass2.filename

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    @pytest.mark.filterwarnings(
        "ignore:invalid value encountered in log:RuntimeWarning"
    )
    def test__init__09(self, array_type):
        input_array = array_type(self.test_data)
        file_name1 = "test_filename1.pickle"
        file_name2 = "test_filename2.pickle"
        KrigingClass1 = KrigingModel(input_array, fname=file_name1, overwrite=True)
        results = KrigingClass1.training()
        KrigingClass2 = KrigingModel(input_array, fname=file_name2, overwrite=True)
        assert KrigingClass1.filename == file_name1
        assert KrigingClass2.filename == file_name2

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_covariance_matrix_generator(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        p = 2
        theta = np.array([1, 2])
        reg_param = 1.00000000e-06
        cov_matrix = KrigingClass.covariance_matrix_generator(
            KrigingClass.x_data_scaled, theta, reg_param, p
        )

        cov_matrix_exp = np.array(
            [
                [1.000001, 0.60653066, 0.13533528],
                [0.60653066, 1.000001, 0.60653066],
                [0.13533528, 0.60653066, 1.000001],
            ]
        )
        np.testing.assert_array_equal(
            np.round(cov_matrix, 7), np.round(cov_matrix_exp, 7)
        )

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_covariance_inverse_generator_01(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        cov_matrix = np.array(
            [
                [1.000001, 0.60653066, 0.13533528],
                [0.60653066, 1.000001, 0.60653066],
                [0.13533528, 0.60653066, 1.000001],
            ]
        )
        cov_matrix_inv_exp = np.array(
            [
                [1.82957788, -1.51792604, 0.67306158],
                [-1.51792604, 2.84133453, -1.51792604],
                [0.67306158, -1.51792604, 1.82957788],
            ]
        )

        inverse_x = KrigingClass.covariance_inverse_generator(cov_matrix)
        np.testing.assert_array_equal(
            np.round(inverse_x, 7), np.round(cov_matrix_inv_exp, 7)
        )

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_covariance_inverse_generator_02(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        cov_matrix = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        inverse_x = KrigingClass.covariance_inverse_generator(cov_matrix)
        np.testing.assert_array_equal(np.round(inverse_x, 7), np.round(cov_matrix, 7))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_kriging_mean(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        cov_matrix_inv = np.array(
            [
                [1.82957788, -1.51792604, 0.67306158],
                [-1.51792604, 2.84133453, -1.51792604],
                [0.67306158, -1.51792604, 1.82957788],
            ]
        )
        kriging_mean = KrigingClass.kriging_mean(cov_matrix_inv, KrigingClass.y_data)
        kriging_mean_exp = 20.18496
        assert np.round(kriging_mean_exp, 5) == np.round(kriging_mean[0][0], 5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_y_mu_calculation(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        kriging_mean = 20.18496
        y_mu = KrigingClass.y_mu_calculation(KrigingClass.y_data, kriging_mean)
        y_mu_exp = np.array([[-18.18496], [-6.93496], [16.81504]])
        np.testing.assert_array_equal(np.round(y_mu, 5), np.round(y_mu_exp, 5))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_kriging_sd(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        cov_matrix_inv = np.array(
            [
                [1.82957788, -1.51792604, 0.67306158],
                [-1.51792604, 2.84133453, -1.51792604],
                [0.67306158, -1.51792604, 1.82957788],
            ]
        )
        y_mu_exp = np.array([[-18.18496], [-6.93496], [16.81504]])
        sigma_sq = KrigingClass.kriging_sd(
            cov_matrix_inv, y_mu_exp, KrigingClass.y_data.shape[0]
        )
        sigma_sq_exp = 272.84104637
        assert np.round(sigma_sq_exp, 5) == np.round(sigma_sq[0][0], 5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_print_fun(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        KrigingClass.print_fun(1, 2, 3.7)
        sys.stdout = sys.__stdout__
        assert "at minimum 2.0000 accepted 3\n" == capturedOutput.getvalue()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_objective_function(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        conc_log_like = KrigingClass.objective_function(
            var_vector, KrigingClass.x_data_scaled, KrigingClass.y_data, p
        )

        conc_log_like_exp = 8.0408619
        assert np.round(conc_log_like_exp, 5) == np.round(conc_log_like, 5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_numerical_gradient_01(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=True)
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        grad_vec = KrigingClass.numerical_gradient(
            var_vector, KrigingClass.x_data_scaled, KrigingClass.y_data, p
        )
        grad_vec_exp = np.array([0, 0, 8.8817842e-10])
        np.testing.assert_array_equal(np.round(grad_vec, 5), np.round(grad_vec_exp, 5))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_numerical_gradient_02(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=False)
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        grad_vec = KrigingClass.numerical_gradient(
            var_vector, KrigingClass.x_data_scaled, KrigingClass.y_data, p
        )
        grad_vec_exp = np.array([0, 0, 0])
        np.testing.assert_array_equal(np.round(grad_vec, 5), np.round(grad_vec_exp, 5))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_parameter_optimization_01(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3])
        p = 2
        np.random.seed(0)
        opt_results = KrigingClass.parameter_optimization(p)
        assert len(opt_results.x) == 3
        assert opt_results.success == True

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_parameter_optimization_02(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], numerical_gradients=False)
        p = 2
        opt_results = KrigingClass.parameter_optimization(p)
        assert len(opt_results.x) == 3
        assert opt_results.minimization_failures == False

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_optimal_parameter_evaluation(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3])
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        (
            theta,
            reg_param,
            mean,
            variance,
            cov_mat,
            cov_inv,
            y_mu,
        ) = KrigingClass.optimal_parameter_evaluation(var_vector, p)
        np.testing.assert_array_equal(theta, [10**1, 10**2])
        np.testing.assert_array_equal(reg_param, 1.00000000e-06)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_error_calculation(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=False)
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        (
            theta,
            reg_param,
            mean,
            variance,
            cov_mat,
            cov_inv,
            y_mu,
        ) = KrigingClass.optimal_parameter_evaluation(var_vector, p)
        y_prediction_exp = np.zeros((KrigingClass.x_data_scaled.shape[0], 1))
        for i in range(0, KrigingClass.x_data_scaled.shape[0]):
            cmt = (
                np.matmul(
                    (
                        (
                            np.abs(
                                KrigingClass.x_data_scaled[i, :]
                                - KrigingClass.x_data_scaled
                            )
                        )
                        ** p
                    ),
                    theta,
                )
            ).transpose()
            cov_matrix_tests = np.exp(-1 * cmt)
            y_prediction_val = mean + np.matmul(
                np.matmul(cov_matrix_tests.transpose(), cov_inv), y_mu
            )
            y_prediction_exp[i, 0] = y_prediction_val.item()

        ss_error, rmse_error, y_prediction = KrigingClass.error_calculation(
            theta,
            p,
            mean,
            cov_inv,
            y_mu,
            KrigingClass.x_data_scaled,
            KrigingClass.y_data,
        )

        np.testing.assert_array_equal(y_prediction, y_prediction_exp)
        assert (
            np.sum((KrigingClass.y_data - y_prediction_exp) ** 2)
            / KrigingClass.x_data_scaled.shape[0]
            == ss_error
        )
        assert (
            np.sqrt(
                np.sum((KrigingClass.y_data - y_prediction_exp) ** 2)
                / KrigingClass.x_data_scaled.shape[0]
            )
            == rmse_error
        )

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_r2_calculation(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=False)
        p = 2
        var_vector = np.array([1, 2, 1.00000000e-06])
        (
            theta,
            reg_param,
            mean,
            variance,
            cov_mat,
            cov_inv,
            y_mu,
        ) = KrigingClass.optimal_parameter_evaluation(var_vector, p)

        ss_error, rmse_error, y_prediction = KrigingClass.error_calculation(
            theta,
            p,
            mean,
            cov_inv,
            y_mu,
            KrigingClass.x_data_scaled,
            KrigingClass.y_data,
        )
        r_square = KrigingClass.r2_calculation(KrigingClass.y_data, y_prediction)
        assert 0.999999999999 == r_square

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    @pytest.mark.filterwarnings(
        "ignore:invalid value encountered in log:RuntimeWarning"
    )
    def test_predict_output_01(self, array_type):
        input_array = array_type(self.training_data)
        np.random.seed(0)
        KrigingClass = KrigingModel(input_array)
        results = KrigingClass.training()
        y_pred = KrigingClass.predict_output(KrigingClass.x_data_scaled)
        assert y_pred.shape[0] == KrigingClass.x_data_scaled.shape[0]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    @pytest.mark.filterwarnings(
        "ignore:invalid value encountered in log:RuntimeWarning"
    )
    def test_predict_output(self, array_type):
        input_array = array_type(self.training_data)
        np.random.seed(0)
        KrigingClass = KrigingModel(input_array)
        results = KrigingClass.training()
        y_pred = KrigingClass.predict_output(np.array([0.1, 0.2]))
        assert y_pred.shape[0] == 1

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_training(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array[0:3], regularization=False)
        np.random.seed(0)
        p = 2
        bh_results = KrigingClass.parameter_optimization(p)
        # Calculate other variables and parameters
        (
            optimal_theta,
            optimal_reg_param,
            optimal_mean,
            optimal_variance,
            optimal_cov_mat,
            opt_cov_inv,
            optimal_ymu,
        ) = KrigingClass.optimal_parameter_evaluation(bh_results.x, p)
        # Training performance
        (
            training_ss_error,
            rmse_error,
            y_training_predictions,
        ) = KrigingClass.error_calculation(
            optimal_theta,
            p,
            optimal_mean,
            opt_cov_inv,
            optimal_ymu,
            KrigingClass.x_data_scaled,
            KrigingClass.y_data,
        )
        r2_training = KrigingClass.r2_calculation(
            KrigingClass.y_data, y_training_predictions
        )

        np.random.seed(0)
        results = KrigingClass.training()
        np.testing.assert_array_equal(results.optimal_weights, optimal_theta)
        np.testing.assert_array_equal(
            results.regularization_parameter, optimal_reg_param
        )
        np.testing.assert_array_equal(results.optimal_mean, optimal_mean)
        np.testing.assert_array_equal(results.optimal_variance, optimal_variance)
        np.testing.assert_array_equal(
            results.optimal_covariance_matrix, optimal_cov_mat
        )
        np.testing.assert_array_equal(results.optimal_y_mu, optimal_ymu)
        np.testing.assert_array_equal(
            results.output_predictions, y_training_predictions
        )
        np.testing.assert_array_equal(results.training_R2, r2_training)
        np.testing.assert_array_equal(results.training_rmse, rmse_error)
        np.testing.assert_array_equal(results.optimal_p, p)
        np.testing.assert_array_equal(results.x_data, KrigingClass.x_data)
        np.testing.assert_array_equal(results.x_data_scaled, KrigingClass.x_data_scaled)
        np.testing.assert_array_equal(results.x_data_min, KrigingClass.x_data_min)
        np.testing.assert_array_equal(results.x_data_max, KrigingClass.x_data_max)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [pd.DataFrame])
    def test_get_feature_vector_01(self, array_type):
        input_array = array_type(self.full_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        p = KrigingClass.get_feature_vector()
        expected_dict = {"x1": 0, "x2": 0}
        assert expected_dict == p.extract_values()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_get_feature_vector_02(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        p = KrigingClass.get_feature_vector()
        expected_dict = {0: 0, 1: 0}
        assert expected_dict == p.extract_values()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_kriging_generate_expression(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        results = KrigingClass.training()
        p = KrigingClass.get_feature_vector()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = results.generate_expression((lv))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_pickle_load01(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        results = KrigingClass.training()
        KrigingClass.pickle_load(KrigingClass.filename)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_pickle_load02(self, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        with pytest.raises(Exception):
            KrigingClass.pickle_load("file_not_existing.pickle")

    @pytest.mark.unit
    @patch("matplotlib.pyplot.show")
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_parity_residual_plots(self, mock_show, array_type):
        input_array = array_type(self.training_data)
        KrigingClass = KrigingModel(input_array, regularization=False)
        results = KrigingClass.training()
        KrigingClass.parity_residual_plots()


if __name__ == "__main__":
    pytest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard IDAES Product block.
"""

# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog

# Product blocks can reuse the Feed initializer
# For consistency and future proofing, import with new name
from idaes.models.unit_models.feed import FeedInitializer as ProductInitializer

__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Product")
class ProductData(UnitModelBlockData):
    """
    Standard Product Block Class
    """

    # Set default initializer
    default_initializer = ProductInitializer

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Product blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Product blocks do not contain holdup, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property
calculations, **default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None

        """
        # Call UnitModel.build to setup dynamics
        super(ProductData, self).build()

        # Check for default property package
        self._get_property_package()

        # Add State Block
        self.properties = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in product",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.property_package_args
        )

        # Add references to all state vars
        s_vars = self.properties[self.flowsheet().time.first()].define_state_vars()
        for s in s_vars:
            l_name = s_vars[s].local_name
            if s_vars[s].is_indexed():
                slicer = self.properties[:].component(l_name)[...]
            else:
                slicer = self.properties[:].component(l_name)

            r = Reference(slicer)
            setattr(self, s, r)

        # Add outlet port
        self.add_port(name="inlet", block=self.properties, doc="Inlet Port")

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        This method calls the initialization method of the state block.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None,
                     use default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        # ---------------------------------------------------------------------
        # Initialize state block
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        blk.properties.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args
        )
        init_log.info("Initialization Complete.")

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {"Inlet": self.inlet}, time_point=time_point
        )



# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions for initialization of IDAES models.
"""

from pyomo.environ import (
    Block,
    check_optimal_termination,
    Constraint,
    value,
)
from pyomo.network import Arc
from pyomo.dae import ContinuousSet
from pyomo.core.expr.visitor import identify_variables

from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.dyn_utils import (
    get_activity_dict,
    deactivate_model_at,
    deactivate_constraints_unindexed_by,
    fix_vars_unindexed_by,
    get_derivatives_at,
    copy_values_at_time,
    get_implicit_index_of_set,
)
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver

__author__ = "Andrew Lee, John Siirola, Robert Parker"


def fix_state_vars(blk, state_args=None):
    """
    Method for fixing state variables within StateBlocks. Method takes an
    optional argument of values to use when fixing variables.

    Args:
        blk : An IDAES StateBlock object in which to fix the state variables
        state_args : a dict containing values to use when fixing state
                variables. Keys must match with names used in the
                define_state_vars method, and indices of any variables must
                agree.

    Returns:
        A dict keyed by block index, state variable name (as defined by
        define_state_variables) and variable index indicating the fixed status
        of each variable before the fix_state_vars method was applied.
    """
    # For sanity, handle cases where state_args is None
    if state_args is None:
        state_args = {}

    flags = {}
    for k in blk.keys():
        for n, v in blk[k].define_state_vars().items():
            for i in v:
                flags[k, n, i] = v[i].is_fixed()

                # If not fixed, fix at either guess provided or current value
                if not v[i].is_fixed():
                    if n in state_args:
                        # Try to get initial guess from state_args
                        try:
                            if i is None:
                                val = state_args[n]
                            else:
                                val = state_args[n][i]
                        except KeyError:
                            raise ConfigurationError(
                                "Indexes in state_args did not agree with "
                                "those of state variable {}. Please ensure "
                                "that indexes for initial guesses are correct.".format(
                                    n
                                )
                            )
                        v[i].fix(val)
                    else:
                        # No guess, try to use current value
                        if v[i].value is not None:
                            v[i].fix()
                        else:
                            # No initial value - raise Exception before this
                            # gets to a solver.
                            raise ConfigurationError(
                                "State variable {} does not have a value "
                                "assigned. This usually occurs when a Var "
                                "is not assigned an initial value when it is "
                                "created. Please ensure all variables have "
                                "valid values before fixing them.".format(v.name)
                            )

    return flags


def revert_state_vars(blk, flags):
    """
    Method to revert the fixed state of the state variables within an IDAES
    StateBlock based on a set of flags of the previous state.

    Args:
        blk : an IDAES StateBlock
        flags : a dict of bools indicating previous state with keys in the form
                (StateBlock index, state variable name (as defined by
                define_state_vars), var indices).

    Returns:
        None
    """
    for k in blk.keys():
        for n, v in blk[k].define_state_vars().items():
            for i in v:
                try:
                    if not flags[k, n, i]:
                        v[i].unfix()
                except KeyError:
                    raise ConfigurationError(
                        "Indices of flags proved do not match with indices of"
                        "the StateBlock. Please make sure you are using the "
                        "correct StateBlock."
                    )


def propagate_state(
    destination=None, source=None, arc=None, direction="forward", overwrite_fixed=False
):
    """
    This method propagates values between Ports along Arcs. Values can be
    propagated in either direction using the direction argument.

    Args:
        destination (Port): Port to copy values to or None if specifying arc
        source (Port): Port to copy values from or None if specifying arc
        arc (Arc): If arc is provided, use arc to define source and destination
        direction (str): Direction in which to propagate values.
                Default = 'forward' Valid values: 'forward', 'backward'.
        overwrite_fixed (bool): If True overwrite fixed values, otherwise do not
            overwrite fixed values.

    Returns:
        None
    """
    _log = idaeslog.getLogger(__name__)
    # Allow an arc to be passed as a positional arg
    if destination is not None and source is None and destination.ctype is Arc:
        arc = destination
        if source in ("forward", "backward"):
            direction = source
        destination = None
        source = None
    # Check that only arc or source and destination are passed
    if arc is None and (destination is None or source is None):
        raise RuntimeError(
            "In propagate_state(), must provide source and destination or arc"
        )
    if arc is not None:
        if destination is not None or source is not None:
            raise RuntimeError(
                "In propagate_state(), provide only arc or source and destination"
            )
        try:
            source = arc.src
            destination = arc.dest
        except AttributeError:
            _log.error(
                "In propagate_state(), unexpected type of arc "
                "argument. Value must be a Pyomo Arc"
            )
            raise

    if direction == "forward":
        pass
    elif direction == "backward":
        source, destination = destination, source
    else:
        raise ValueError(
            "Unexpected value for direction argument: ({}). "
            "Value must be either 'forward' or 'backward'.".format(direction)
        )

    try:
        destination_vars = destination.vars
    except AttributeError:
        _log.error(
            "In propagate_state(), unexpected type of destination "
            "port argument. Value must be a Pyomo Port"
        )
        raise
    try:
        source_vars = source.vars
    except AttributeError:
        _log.error(
            "In propagate_state(), unexpected type of destination "
            "source argument. Value must be a Pyomo Port"
        )
        raise

    # Copy values
    for k, v in destination_vars.items():
        try:
            for i in v:
                if v[i].is_variable_type() and ((not v[i].fixed) or overwrite_fixed):
                    v[i].value = value(source_vars[k][i])
                elif not v[i].is_variable_type():
                    raise TypeError(
                        f"propagate_state() is can only change the value of "
                        f"variables and cannot set a {v[i].ctype}.  This "
                        f"likely indicates either a malformed port or a misuse "
                        f"of propagate_state."
                    )
        except KeyError as e:
            raise KeyError(
                "In propagate_state, variables have incompatible index sets"
            ) from e


# HACK, courtesy of J. Siirola
def solve_indexed_blocks(solver, blocks, **kwds):
    """
    This method allows for solving of Indexed Block components as if they were
    a single Block. A temporary Block object is created which is populated with
    the contents of the objects in the blocks argument and then solved.

    Args:
        solver : a Pyomo solver object to use when solving the Indexed Block
        blocks : an object which inherits from Block, or a list of Blocks
        kwds : a dict of arguments to be passed to the solver

    Returns:
        A Pyomo solver results object
    """
    # We need to play with Pyomo internals for this
    # Check blocks argument, and convert to a list of Blocks
    if isinstance(blocks, Block):
        blocks = [blocks]

    try:
        # Create a temporary Block
        tmp = Block(concrete=True)

        nBlocks = len(blocks)

        # Iterate over indexed objects
        for i, b in enumerate(blocks):
            # Check that object is a Block
            if not isinstance(b, Block):
                raise TypeError(
                    "Trying to apply solve_indexed_blocks to "
                    "object containing non-Block objects"
                )
            # Append components of BlockData to temporary Block
            try:
                tmp._decl["block_%s" % i] = i  # pylint: disable=protected-access
                tmp._decl_order.append(  # pylint: disable=protected-access
                    (b, i + 1 if i < nBlocks - 1 else None)
                )
            except Exception:
                # PYLINT-TODO
                # pylint: disable-next=broad-exception-raised
                raise Exception(
                    "solve_indexed_blocks method failed adding "
                    "components to temporary block."
                )

        # Set ctypes on temporary Block
        tmp._ctypes[Block] = [  # pylint: disable=protected-access
            0,
            nBlocks - 1,
            nBlocks,
        ]

        # Solve temporary Block
        results = solver.solve(tmp, **kwds)

    finally:
        # Clean up temporary Block contents so they are not removed when Block
        # is garbage collected.
        tmp._decl = {}  # pylint: disable=protected-access
        tmp._decl_order = []  # pylint: disable=protected-access
        tmp._ctypes = {}  # pylint: disable=protected-access

    # Return results
    return results


def initialize_by_time_element(fs, time, **kwargs):
    """
    Function to initialize Flowsheet fs element-by-element along
    ContinuousSet time. Assumes sufficient initialization/correct degrees
    of freedom such that the first finite element can be solved immediately
    and each subsequent finite element can be solved by fixing differential
    and derivative variables at the initial time point of that finite element.

    Args:
        fs : Flowsheet to initialize
        time : Set whose elements will be solved for individually
        solver : Pyomo solver object initialized with user's desired options
        outlvl : IDAES logger outlvl
        ignore_dof : Bool. If True, checks for square problems will be skipped.

    Returns:
        None
    """
    if not fs.is_flowsheet():
        raise TypeError("First arg must be a FlowsheetBlock")
    if not isinstance(time, ContinuousSet):
        raise TypeError("Second arg must be a ContinuousSet")

    if time.get_discretization_info() == {}:
        raise ValueError("ContinuousSet must be discretized")

    scheme = time.get_discretization_info()["scheme"]
    nfe = time.get_discretization_info()["nfe"]

    if scheme == "LAGRANGE-RADAU":
        ncp = time.get_discretization_info()["ncp"]
    elif scheme == "LAGRANGE-LEGENDRE":
        raise NotImplementedError(
            "Initialization does not support collocation with Legendre roots"
        )
    elif scheme == "BACKWARD Difference":
        ncp = 1
    elif scheme == "FORWARD Difference":
        ncp = 1
        raise NotImplementedError(
            "Forward initialization (explicit Euler) has not yet been implemented"
        )
    elif scheme == "CENTRAL Difference":
        raise NotImplementedError(
            "Initialization does not support central finite difference"
        )
    else:
        raise ValueError(
            "Unrecognized discretization scheme. "
            "Has the model been discretized along the provided ContinuousSet?"
        )
    # Disallow Central/Legendre discretizations.
    # Neither of these seem to be square by default for multi-finite element
    # initial value problems.

    # Create logger objects
    outlvl = kwargs.pop("outlvl", idaeslog.NOTSET)
    init_log = idaeslog.getInitLogger(__name__, level=outlvl)
    solver_log = idaeslog.getSolveLogger(__name__, level=outlvl)

    ignore_dof = kwargs.pop("ignore_dof", False)
    solver = kwargs.pop("solver", get_solver("ipopt_v2"))
    fix_diff_only = kwargs.pop("fix_diff_only", True)
    # This option makes the assumption that the only variables that
    # link constraints to previous points in time (which must be fixed)
    # are the derivatives and differential variables. Not true if a controller
    # is being present, but should be a good assumption otherwise, and is
    # significantly faster than searching each constraint for time-linking
    # variables.

    if not ignore_dof:
        if degrees_of_freedom(fs) != 0:
            msg = (
                "Original model has nonzero degrees of freedom. This was "
                "unexpected. Use keyword arg ignore_dof=True to skip this "
                "check."
            )
            init_log.error(msg)
            raise ValueError("Nonzero degrees of freedom.")

    # Get dict telling which constraints/blocks are already inactive:
    # dict: id(compdata) -> bool (is active?)
    was_originally_active = get_activity_dict(fs)

    # Deactivate flowsheet except at t0, solve to ensure consistency
    # of initial conditions.
    non_initial_time = [t for t in time]
    non_initial_time.remove(time.first())
    deactivated = deactivate_model_at(fs, time, non_initial_time, outlvl=idaeslog.ERROR)

    if not ignore_dof:
        if degrees_of_freedom(fs) != 0:
            msg = (
                "Model has nonzero degrees of freedom at initial conditions."
                " This was unexpected. Use keyword arg ignore_dof=True to skip"
                " this check."
            )
            init_log.error(msg)
            raise ValueError("Nonzero degrees of freedom.")

    init_log.info(
        "Model is inactive except at t=0. Solving for consistent initial conditions."
    )
    with idaeslog.solver_log(solver_log, level=idaeslog.DEBUG) as slc:
        results = solver.solve(fs, tee=slc.tee)
    if check_optimal_termination(results):
        init_log.info("Successfully solved for consistent initial conditions")
    else:
        init_log.error("Failed to solve for consistent initial conditions")
        raise ValueError("Solver failed in initialization")

    deactivated[time.first()] = deactivate_model_at(
        fs, time, time.first(), outlvl=idaeslog.ERROR
    )[time.first()]

    # Here, deactivate non-time-indexed components. Do this after solve
    # for initial conditions in case these were used to specify initial
    # conditions
    con_unindexed_by_time = deactivate_constraints_unindexed_by(fs, time)
    var_unindexed_by_time = fix_vars_unindexed_by(fs, time)

    # Now model is completely inactive

    # For each timestep, we need to
    # 1. Activate model at points we're solving for
    # 2. Fix initial conditions (differential variables at previous timestep)
    #    of finite element
    # 3. Solve the (now) square system
    # 4. Revert the model to its prior state

    # This will make use of the following dictionaries mapping
    # time points -> time derivatives and time-differential variables
    derivs_at_time = get_derivatives_at(fs, time, [t for t in time])
    dvars_at_time = {
        t: [d.parent_component().get_state_var()[d.index()] for d in derivs_at_time[t]]
        for t in time
    }

    # Perform a solve for 1 -> nfe; i is the index of the finite element
    init_log.info(
        "Flowsheet has been deactivated. Beginning element-wise initialization"
    )
    for i in range(1, nfe + 1):
        t_prev = time.at((i - 1) * ncp + 1)
        # Non-initial time points in the finite element:
        fe = [time.at(k) for k in range((i - 1) * ncp + 2, i * ncp + 2)]

        init_log.info(f"Entering step {i}/{nfe} of initialization")

        # Activate components of model that were active in the presumably
        # square original system
        for t in fe:
            for comp in deactivated[t]:
                if was_originally_active[id(comp)]:
                    comp.activate()

        # Get lists of derivative and differential variables
        # at initial time point of finite element
        init_deriv_list = derivs_at_time[t_prev]
        init_dvar_list = dvars_at_time[t_prev]

        # Variables that were originally fixed
        fixed_vars = []
        if fix_diff_only:
            for drv in init_deriv_list:
                # Cannot fix variables with value None.
                # Any variable with value None was not solved for
                # (either stale or not included in previous solve)
                # and we don't want to fix it.
                if not drv.fixed:
                    fixed_vars.append(drv)
                if not drv.value is None:
                    drv.fix()
            for dv in init_dvar_list:
                if not dv.fixed:
                    fixed_vars.append(dv)
                if not dv.value is None:
                    dv.fix()
        else:
            for con in fs.component_data_objects(Constraint, active=True):
                for var in identify_variables(con.expr, include_fixed=False):
                    t_idx = get_implicit_index_of_set(var, time)
                    if t_idx is None:
                        continue
                    if t_idx <= t_prev:
                        fixed_vars.append(var)
                        var.fix()

        # Initialize finite element from its initial conditions
        for t in fe:
            copy_values_at_time(
                fs, fs, t, t_prev, copy_fixed=False, outlvl=idaeslog.ERROR
            )

        # Log that we are solving finite element {i}
        init_log.info(f"Solving finite element {i}")

        if not ignore_dof:
            if degrees_of_freedom(fs) != 0:
                msg = (
                    f"Model has nonzero degrees of freedom at finite element"
                    f" {i}. This was unexpected. "
                    f"Use keyword arg ignore_dof=True to skip this check."
                )
                init_log.error(msg)
                raise ValueError("Nonzero degrees of freedom")

        with idaeslog.solver_log(solver_log, level=idaeslog.DEBUG) as slc:
            results = solver.solve(fs, tee=slc.tee)
        if check_optimal_termination(results):
            init_log.info(f"Successfully solved finite element {i}")
        else:
            init_log.error(f"Failed to solve finite element {i}")
            raise ValueError("Failure in initialization solve")

        # Deactivate components that may have been activated
        for t in fe:
            for comp in deactivated[t]:
                comp.deactivate()

        # Unfix variables that have been fixed
        for var in fixed_vars:
            var.unfix()

        # Log that initialization step {i} has been finished
        init_log.info(f"Initialization step {i} complete")

    # Reactivate components of the model that were originally active
    for t in time:
        for comp in deactivated[t]:
            if was_originally_active[id(comp)]:
                comp.activate()

    for con in con_unindexed_by_time:
        con.activate()
    for var in var_unindexed_by_time:
        var.unfix()

    # Logger message that initialization is finished
    init_log.info("Initialization completed. Model has been reactivated")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Wassiljew-Mason-Saxena rules to obtain mixture thermal conductivity of a low pressure
gas from component thermal conductivities. Has a synergy with the Wilke viscosity
mixing rules because they both rely on the same phi_ij terms.

Method taken from Properties of Gases and Liquids, 5th Ed., Sections 10-6-1 and 10-6-2
"""

from idaes.models.properties.modular_properties.transport_properties.viscosity_wilke import (
    ViscosityWilke,
)


class ThermalConductivityWMS(object):
    """Class to use the Wassiljew-Mason-Saxena rules to obtain mixture thermal conductivity"""

    class therm_cond_phase(object):
        """Implement gas mixture thermal conductivity using Wassiljew-Mason-Saxena rules"""

        @staticmethod
        def build_parameters(pobj):
            """The parameters needed by this method are the same as the ViscosityWilke method"""
            ViscosityWilke.build_parameters(pobj)

        @staticmethod
        def return_expression(b, p):
            """Return expression for therm_cond_phase"""
            # Properties of Gases and Liquids, Eq. 9-5.14
            # and Eq. 10-6.4
            ViscosityWilke.build_phi_ij(b, p)
            if not hasattr(b, "_therm_cond_phase_comp"):
                b._make_therm_cond_phase_comp()  # pylint: disable=protected-access

            # Properties of Gases and Liquids, Eq. 10-6.2
            return sum(
                [
                    b.mole_frac_phase_comp[p, i]
                    * b._therm_cond_phase_comp[p, i]  # pylint: disable=protected-access
                    / sum(
                        [
                            b.mole_frac_phase_comp[p, j] * b.visc_d_phi_ij[i, j]
                            for j in b.components_in_phase(p)
                        ]
                    )
                    for i in b.components_in_phase(p)
                ]
            )
